import os
import re
from pathlib import Path
from typing import TypedDict

from html_sanitizer import Sanitizer

MARKSAFE = Sanitizer(  # Only allow html tag: <mark>; see restrictive https://github.com/matthiask/html-sanitizer/
    {
        "tags": ("mark",),
        "attributes": {},
        "empty": set(),
        "separate": set(),
    }
)


def set_sql_path(entity: str, source: str) -> str:
    """The sql statement is dependent on the search filters of `entity` and
    `source`.

    If an entity needs full-text-search via sqlite, it needs to have an
    appropriate `fts` table. The querying of this table is through an
    sql statement found in the template: `/templates/<entity_name>/fts/<source>/details.sql`.

    `source` refers to the criteria to use for the search proper, e.g. in searching for
    the phrase `bureau of internal revenue` in the entity Decisions, I can choose to source
    my prospective results from two sources: (a) `body_text` or (b) `case_title`; each of these
    sources will need to have their own subfolder with at least 2 files:

    1. an `ids.sql`: which outputs rowids that match the search phrase; and
    2. a `details.sql`: which outputs the descriptive columns matching the results of the `ids.sql`

    The `ids.sql` file is a Jinja macro to the `details.sql` statement so only the `details.sql` statement
    will be needed.
    """
    return f"{entity}/fts/{source}/details.sql"


class ResultPagination(TypedDict):
    """Each `ResultCollection` can be paginated.

    The data displayed here will affect the buttons to be displayed in
    the template. The start_row and end_row numbers allows template to
    say "Display rows from X to Y". Can join this with the `count`
    parameter from `ResultCollection` to get "Display rows from X to Y
    of `count`" where `count` is the maximum number of rows outside of
    the windowed, paginated collection.
    """

    start_row: int  # the `results` from ResultCollection are based on a start row
    end_row: int  # the `results` from ResultCollection are based on an end row
    next_url: str | None  # if absent, no next button should be displayed on template
    prev_url: str | None  # if absent, no previous button should be displayed on template


class ResultSetup(TypedDict):
    max_count: int
    results: list[dict]
    btn_text: str
    btn_url: str


class SearchFilterResults(TypedDict):
    item: str
    slug: str
    units: list[dict]


class SearchFilterSetup(TypedDict):
    entities: SearchFilterResults
    sources: SearchFilterResults
    actives: list[str]


class SortFilterSetup(TypedDict):
    sort_url_desc: str
    sort_url_asc: str


class ResultItem(TypedDict):
    kind: str
    url: str
    title: str  # what appears in the headline
    blurb: str  # what appears in the description, with <mark> html allowed
    date: str


def shorten_item(text: str):
    """Replace the item's long form keyword with its shorthand format, if
    possible."""
    if "Republic Act" in text:
        return re.sub(r"Republic\s+Act\s+No\.", "RA", text)
    if "Presidential Decree" in text:
        return re.sub(r"Presidential\s+Decree\s+No\.", "PD", text)
    if "Batas Pambansa" in text:
        return re.sub(r"Batas\s+Pambansa\s+Blg\.", "BP", text)
    if "Executive Order" in text:
        return re.sub(r"Executive\s+Order\s+No\.", "EO", text)
    if "Commonwealth Act" in text:
        return re.sub(r"Commonwealth\s+Act\s+No\.", "CA", text)
    if "Administrative Matter" in text:
        return re.sub(r"Administrative\s+Matter\s+No\.", "AM", text)
    if "Bar Matter" in text:
        return re.sub(r"Bar\s+Matter\s+No\.", "BM", text)
    if "Administrative Circular" in text:
        return re.sub(r"Administrative\s+Circular\s+No\.", "AC", text)
    if "Paragraph" in text:
        return re.sub(r"Paragraph", "Par.", text)
    if "Article" in text:
        return re.sub(r"Article", "Art.", text)
    if "Section" in text:
        return re.sub(r"Section", "Sec.", text)
    if "Book" in text:
        return re.sub(r"Book", "Bk.", text)
    if "Chapter" in text:
        return re.sub(r"Chapter", "Ch.", text)
    if "Sub-Container" in text:
        return re.sub(r"Sub-Container(\s+\d+)?", "", text)
    if "Container" in text:
        return re.sub(r"Container(\s+\d+)?", "", text)
    return text


def set_result_item(row: dict, kind: str) -> ResultItem:
    """Each row consists of fields that will be useful in populating each item
    in the search result page.

    This implies that each row emanating from `details.sql` should have
    the following columns: `max_row`, `pk`, `row_date`, `title`, and
    `blurb`.
    """

    # construct url, using the kind of entity and the row `pk`
    url = f"/ent/{kind}/{row['pk']}"

    # the title should be shortened if the kind of title is a statute
    title = shorten_item(row["title"]) if kind == "statutes" else row["title"]

    # the blurb will contain a <mark> tag based on the fts search on sqlite and this should be the only html tag allowed
    blurb = "Undetected snippet."
    if row.get("blurb"):
        blurb = MARKSAFE.sanitize(row["blurb"])

    # the date of the resulting row
    date = row["row_date"]

    return ResultItem(kind=kind, url=url, title=title, blurb=blurb, date=date)
