import os
import re
import sqlite3
from contextlib import closing
from datetime import datetime
from pathlib import Path
from sqlite3 import Connection, OperationalError
from typing import Iterator, NamedTuple

from dotenv import find_dotenv, load_dotenv
from jinja2 import Environment, PackageLoader, select_autoescape
from rich import print

load_dotenv(find_dotenv())


JINJA_ENVIRONMENT = Environment(
    loader=PackageLoader("corpus_sql"),
    autoescape=select_autoescape(),
)


def get_db_path() -> Connection:
    """Gets the proper database connection depending on the existence of
    environment variables."""
    if folder := os.getenv("LOCAL_PATH"):
        p = Path().home() / folder
        conn = sqlite3.connect(p)
        conn.row_factory = sqlite3.Row
        return conn
    if volume := os.getenv("DB_FILE"):
        return sqlite3.connect(volume)
    raise Exception("Not connected to volume nor local file.")


def get_rows(
    p: str, params: dict = {}, show_sql: bool = False
) -> Iterator[dict]:
    """Convert sqlite3 row objects into dictionaries."""
    sql_stmt = JINJA_ENVIRONMENT.get_template(p).render(params)
    with closing(get_db_path()) as conn:  # db needs to be open
        with closing(conn.cursor()) as cursor:
            try:
                if show_sql:
                    print(sql_stmt)
                if cur := cursor.execute(sql_stmt):
                    if rows := cur.fetchall():
                        for row in rows:
                            yield {k: row[k] for k in row.keys()}
            except OperationalError as e:
                raise Exception(f"Could not fetch rows, {e=}")


def get_row(p: str, params: dict = {}, show_sql: bool = False) -> dict:
    """Get the first dict from `get_rows()`"""
    try:
        return next(get_rows(p, params, show_sql))
    except StopIteration:
        raise Exception(f"Could not get {p=} with {params=}")


def get_year(value: str, short: bool = False, format: str = "%Y-%m-%d") -> str:
    """The date format is the default from sql result queries; if short is
    enabled only the last 2 digits of the year are included."""
    date_obj = datetime.strptime(value, format)
    year = date_obj.strftime("%Y")
    return year if not short else f"'{year[-2:]}"


def readable_date(value: str, format: str = "%Y-%m-%d") -> str:
    """The date format is the default from sql result queries; if short is
    enabled only the last 2 digits of the year are included."""
    date_obj = datetime.strptime(value, format)
    readable = date_obj.strftime("%B %-d, %Y")
    return readable


def make_unique_titles(value: str) -> str:
    """The titles may consist of duplicate text since the official title might
    also be the short title, etc."""
    texts = value.split("; ")
    uniques = set(texts)
    return "; ".join(uniques)


def valid_mp(candidate_unit_id: str) -> str:
    """Each unit id is material path following the format: 1.1.1.1.

    etc. where all root nodes start with 1. and all branches / leaves
    terminate in the same fashion, e.g. 1.2.13.16.

    The root's identifier `materialized_path` is always 1. and the value of the specific branch can be determined through the addition of other identifier's to the root. e.g. '1.1.1.2' of 'mv-2022-ra-386-v1' will yield all material nodes of the tree from the root to the specified `mp` branch.
    """
    mp_pattern = re.compile(r"(\d+\.)+")
    if mp_pattern.fullmatch(candidate_unit_id):
        return candidate_unit_id
    raise Exception(f"Bad {candidate_unit_id=}")


class DecisionJustice(NamedTuple):
    per_curiam: bool
    last_name: str | None
    start_term: str | None
    chief_date: str | None


class DecisionIndicator(NamedTuple):
    title: str
    fallo: str
    tags: list[str]


class DecisionCitation(NamedTuple):
    phil: str | None
    scra: str | None
    offg: str | None
    docket: str | None

    def __str__(self):
        l = (self.phil, self.scra, self.offg, self.docket)
        texts = [i for i in l if i]
        return ", ".join(texts)
