import json
from dataclasses import asdict
from typing import Any, Iterator

from .etc import (
    DecisionCitation,
    DecisionIndicator,
    DecisionJustice,
    get_row,
    get_rows,
    valid_mp,
)
from .models import (
    AuthorDetail,
    CodificationDetail,
    DecisionDetail,
    DocumentDetail,
    StatuteDetail,
    UnitDetail,
)


class Commons:
    @classmethod
    def set_url(cls, pk: str) -> dict:
        """Based on the inheriting class, create an absolute url."""
        kind = cls.__name__.lower()
        return dict(kind=kind, url=f"/ent/{kind}/{pk}")

    @classmethod
    def set_urls(cls, rows: list[dict]) -> list[dict]:
        """Update all dict rows, having a `pk` key with `set_url()`, based on
        the inheriting class."""
        return [r | cls.set_url(r["pk"]) for r in rows if r.get("pk")]

    def opt_extract(
        self,
        key: str,
        data: dict,
        allow: bool = True,
        obj=None,
        _type: type = list,
    ) -> tuple[dict, Any]:
        """If `allow` is `True`, a value matching the `key` is removed from the
        `data` dict, decoded to json and returned as a separated value.

        If `allow` is `False`, then the returned value will be `None`. This enables
        a mechanism to exclude a field from being displayed.
        """
        extracted = data.pop(key, None)
        if allow and extracted:
            obj = json.loads(extracted)
            if isinstance(obj, _type):
                return data, obj
            else:
                raise Exception(f"Bad {_type=} for {data[key]}")
        return data, obj


class Decisions(Commons):
    def get_detail(
        self,
        pk: str,
        show_body: bool = True,
        with_statutes: bool = True,
        with_decisions: bool = True,
        snippet: str = None,
    ) -> DecisionDetail:
        params = {"case_pk": f"'{pk}'", "show_body": show_body}
        row = get_row("decisions/detail.sql", params)
        if row.get("ponencia") and not show_body:
            row["ponencia"] = None
        row, indicators = self.opt_extract("indicator_data", row, _type=dict)
        row, citations = self.opt_extract("citation_data", row, _type=dict)
        row, writer = self.opt_extract("justice_data", row, _type=dict)
        row, s = self.opt_extract("statutes", row, with_statutes)
        row, d = self.opt_extract("decisions", row, with_decisions)
        return DecisionDetail(
            **row,
            **self.set_url(pk),
            citation=DecisionCitation(**citations) if citations else None,
            writer=DecisionJustice(**writer) if writer else None,
            indicators=DecisionIndicator(**indicators),
            statutes_cited=Statutes.set_urls(s) if s else [],
            decisions_cited=Decisions.set_urls(d) if d else [],
            snippet=snippet,
        )

    def find_decisions(self, query: str):
        """Convert the querying string to an FTS-matchable string and then extract `query`-matching decisions with snippets."""
        params = {"query": f"""'{query}'"""}
        try:
            if rows := get_rows("decisions/find_decisions.sql", params):
                for row in rows:
                    yield self.get_detail(
                        pk=row["case_pk"],
                        show_body=False,
                        with_statutes=False,
                        with_decisions=False,
                        snippet=row["snippet"],
                    )
        except:
            ...
        return None


class Codifications(Commons):
    def get_detail(
        self,
        pk: str,
        with_units: bool = True,
        with_statutes: bool = True,
    ) -> CodificationDetail:
        params = {"code_pk": f"'{pk}'", "with_units": with_units}
        row = get_row("codifications/detail.sql", params)
        row, s = self.opt_extract("statutes_included", row, with_statutes)
        row, u = self.opt_extract("units", row, with_units)
        return CodificationDetail(
            **row,
            **self.set_url(pk),
            statutes_included=Statutes.set_urls(s) if s else [],
            units=u[0]["units"] if u else [],  # exclude node 1.
        )

    def get_unit(
        self, pk: str, mp: str, with_crumbs: bool = True
    ) -> UnitDetail:
        base = self.get_detail(pk, with_units=False, with_statutes=False)
        return UnitDetail(base, mp, with_crumbs)


class Statutes(Commons):
    def get_detail(
        self,
        pk: str,
        with_units: bool = True,
        get_annotated: bool = True,
        get_related: bool = True,
    ) -> StatuteDetail:
        params = {"statute_pk": f"'{pk}'", "with_units": with_units}
        row = get_row("statutes/detail.sql", params)
        row, t = self.opt_extract("titles", row, _type=dict)
        row, u = self.opt_extract("units", row, with_units)
        row, a = self.opt_extract("annotations", row, get_annotated)
        row, r = self.opt_extract("relations", row, get_related)
        return StatuteDetail(
            **row,
            **self.set_url(pk),
            titles=t,
            units=u[0]["units"] if u else [],
            annotations=Codifications.set_urls(a) if a else [],
            relations=Codifications.set_urls(r) if r else [],
        )

    def get_unit(
        self, pk: str, mp: str, with_crumbs: bool = True
    ) -> UnitDetail:
        base = self.get_detail(
            pk, with_units=False, get_annotated=False, get_related=False
        )
        return UnitDetail(base, mp, with_crumbs)

    def find_unit(self, data: dict) -> dict:
        """The params are ultimately sourced from the codification yaml file
        which contains one or many history nodes.

        Each history node can be associated with a given statute.
        However, there are instances were a simple lookup will not work
        since the unique keys are not well defined. This function looks
        at a customized variant id; if none is available, then the first
        statute serial text that matches.
        """
        statute_keys = ["statute", "date", "variant"]
        unit_keys = ["locator", "caption", "content"]
        keys = statute_keys + unit_keys
        d = parametize_cols(keys, data)
        row = get_row("statutes/find_unit.sql", d)
        return row


class Documents(Commons):
    def get_detail(
        self,
        pk: str,
        with_units: bool = True,
    ) -> DocumentDetail:
        params = {"doc_pk": f"'{pk}'", "with_units": with_units}
        row = get_row("documents/detail.sql", params)
        row, u = self.opt_extract("units", row, with_units)
        return DocumentDetail(
            **row,
            **self.set_url(pk),
            units=u[0]["units"] if u else [],
        )

    def get_unit(
        self, pk: str, mp: str, with_crumbs: bool = True
    ) -> UnitDetail:
        base = self.get_detail(pk, with_units=False)
        return UnitDetail(base, mp, with_crumbs)


class Sources:
    def extract_sources(self, rows: Iterator[dict]) -> tuple:
        unit_objs = []
        case_objs = []
        stat = Statutes()
        case = Decisions()
        for row in rows:
            src = json.loads(row["basis"])
            if "statute" in src and "locator" in src:
                s = stat.find_unit(src)
                if "statute_id" in s and "unit_id" in s:
                    unit_obj = stat.get_unit(
                        pk=s["statute_id"],
                        mp=s["unit_id"],
                        with_crumbs=False,
                    )
                    unit_objs.append(unit_obj)

            if "query" in src:
                q = src["query"]
                if cases := case.find_decisions(q):
                    for case in cases:
                        case_objs.append(case)

        return unit_objs, case_objs

    def list_sources(self, doc_pk: str, mp: str) -> dict:
        """Fetch sources (and their nodes) from a specific unit of a Document
        identified by its material path."""
        params = {"doc_pk": f"'{doc_pk}'", "mp": f"'{valid_mp(mp)}'"}
        rows = get_rows("sources/unit_doc_list.sql", params)
        units, cases = self.extract_sources(rows)
        return {
            "units": [asdict(u) for u in units],
            "cases": [asdict(c) for c in cases],
        }


class Events:
    def list_events(self, code_pk: str, mp: str) -> Iterator[dict]:
        """Fetch events (and their nodes) from a specific unit of a
        Codification identified by its material path."""
        params = {"code_pk": f"'{code_pk}'", "mp": f"'{valid_mp(mp)}'"}
        rows = get_rows("events/unit_code_list.sql", params)
        stat = Statutes()
        for row in rows:
            row["basis"] = json.loads(row["basis"])
            if "statute" in row["basis"]:
                found = stat.find_unit(row["basis"])
                row |= found
            yield row


class Authors:
    def detail(self, email: str):
        stmt = "authors/detail.sql"
        params = {"email": f"'{email}'"}
        row = get_row(stmt, params)
        return AuthorDetail(**row)


def parametize_cols(keys: list[str], data: dict) -> dict:
    """Two operations done during clean:

    1. Return a limited subset of the `data`
    2. Of the subset returned, converts empty string values e.g. {'x':''}  into None, e.g. {'x': None}.

    Useful for parameterized queries to ensure that pugsql can properly bind values, even None ones to the sql statement, e.g.

    e.g. to `find_statute_unit()`, possible data is:

    {
        "locator": "Article 2",
        "caption": '',
        "content": "Laws shall take effect",
        "statute": "Republic Act No. 386",
        "date": '',
        "variant": '',
    }

    our required fields are:

    {
        "locator": "Article 2",
        "caption": '',
        "content": "Laws shall take effect",
    }

    since `caption` is an empty string `''`, this must be converted to `NULL`:

    {
        "locator": "Article 2",
        "caption": `NULL`,
        "content": "Laws shall take effect",
    }

    so that when `/event/_unit_id_from_data.sql` is referenced, `{{caption}}` can evaluate to sql's `NULL`.
    """
    for key in keys:
        if raw := data.get(key) and data[key].strip():
            data[key] = f"'{raw.strip()}'"
        else:
            data[key] = "NULL"
    return data
