import datetime
import json
from dataclasses import dataclass, field
from typing import Iterator

from corpus_sql.etc import (
    DecisionCitation,
    DecisionIndicator,
    DecisionJustice,
    get_rows,
    shorten_item,
    valid_mp,
)


@dataclass
class DecisionDetail:
    pk: str
    kind: str
    url: str
    source: str
    date_prom: str
    category: str
    composition: str
    voting_line: str
    ponencia: str | None = None
    citation: DecisionCitation | None = None
    writer: DecisionJustice | None = None
    indicators: DecisionIndicator | None = None
    statutes_cited: list[dict] = field(default_factory=list)
    decisions_cited: list[dict] = field(default_factory=list)
    snippet: str | None = None

    def __repr__(self) -> str:
        return f"{self.indicators.title if self.indicators else self.pk}, {self.date_prom}"

    @property
    def citation_text(self) -> str | None:
        return str(self.citation)

    @property
    def get_justice_label(self):
        """Create a justice string; when a `chief_date` exists for a riter and
        it is less than the `decision_date`, add `C.J.`; otherwise, just
        `J.`."""
        if not self.writer:
            return None

        is_chief = False
        if chief_date := self.writer.chief_date:
            is_chief = datetime.strptime(
                chief_date, "%Y-%m-%d"
            ) < datetime.strptime(self.date_prom, "%Y-%m-%d")

        if last_name := self.writer.last_name:
            return f"{last_name}, {'C.' if is_chief else ''}J."
        elif self.writer.per_curiam:
            return "Per Curiam"
        return "Could not detect."


@dataclass
class DocumentDetail:
    pk: str
    kind: str
    url: str
    version: str
    title: str
    description: str
    publication_date: str
    authors: list[dict] = field(default_factory=list)
    units: list[dict] = field(default_factory=list)

    def __repr__(self) -> str:
        return f"Document: {self.title}, {self.publication_date}"


@dataclass
class CodificationDetail:
    pk: str
    kind: str
    url: str
    version: str
    title: str
    description: str
    publication_date: str
    statute_id: str
    base_serial_title: str
    base_official_title: str
    base_st_date: str
    authors: list[dict] = field(default_factory=list)
    statutes_included: list[dict] = field(default_factory=list)
    units: list[dict] = field(default_factory=list)

    def __repr__(self) -> str:
        return f"Annotated {self.base_serial_title}, {self.publication_date}"


@dataclass
class StatuteDetail:
    pk: str
    kind: str
    url: str
    category: str
    identifier: str
    promulgation_date: str
    titles: dict
    annotations: list[dict] = field(default_factory=list)
    relations: list[dict] = field(default_factory=list)
    units: list[dict] = field(default_factory=list)

    def __repr__(self) -> str:
        return f"{self.titles['serial']}, {self.promulgation_date}"


@dataclass
class UnitDetail:
    base: StatuteDetail | CodificationDetail | DocumentDetail
    mp: str
    with_crumbs: bool = True
    copy_text: str | None = None
    crumbs: list[dict] = field(default_factory=list)
    subtree: list[dict] = field(default_factory=list)

    def __repr__(self) -> str:
        return f"{self.copy_text}. ID: {self.mp} of {self.base}"

    def __post_init__(self):
        self.nodes = list(self.get_unit_nodes())
        if self.with_crumbs:
            self.crumbs, self.copy_text = self.set_crumbs_and_copy()
        self.subtree = self.build_subtree()

    def get_unit_nodes(self) -> Iterator[dict]:
        stmt = "_utils/material_path.sql"
        params = {
            "table": self.base.kind.capitalize(),
            "table_pk": f"'{self.base.pk}'",
            "unit_mp": f"'{valid_mp(self.mp)}'",
        }
        return (json.loads(r["node_data"]) for r in get_rows(stmt, params))

    def set_crumbs_and_copy(self) -> tuple[list[dict], str]:
        """The crumbs are reversed to create the copy."""

        def get_copies(rows: list[dict]):
            for c in rows:
                if c.get("label") and c.get("caption"):
                    if c["label"] == c["caption"]:  # RA386 Preliminary Title
                        yield c["label"]
                    else:
                        yield f"{c['label']} ({c['caption']})"  # Ch. 1 (Effect and Application of Laws)
                elif c.get("label"):
                    yield c["label"]

        crumbs = list(self.extract_crumbs())
        crumbs[0]["url"] = self.base.url
        return crumbs, ", ".join(reversed(list(get_copies(crumbs))))

    def extract_crumbs(self) -> Iterator[dict]:
        """The subtree's position is recorded through breadcrumbs (`crumbs`);
        and each crumb can later be assigned a url."""

        def isolate_nodes(rows: list[dict], keys: list[str]) -> Iterator[dict]:
            for row in rows:
                data: dict[str, str] = {}
                for k, v in row.items():
                    if k in keys:
                        data |= {k: v}
                yield data

        for node in isolate_nodes(self.nodes, ["id", "item", "caption"]):
            url = f"/unit/{self.base.kind}/{self.base.pk}/{node['id']}"
            if raw_item := node.get("item", None).strip():
                if item := shorten_item(raw_item):
                    yield node | {"label": item, "url": url}
                else:  # e.g. Container items = null, so use caption instead.
                    if paired := node.get("caption", "").strip():
                        yield node | {"label": paired, "url": url}
            elif solo := node.get("caption", "").strip():
                yield node | {"label": solo, "url": url}

    def build_subtree(self):
        """Construct a subtree, i.e. a branch, provided a hierarchical order of
        rows: root node (as the first row) to the deepest leaf node of the
        materialized path (last row).
        """

        def add_child_units(node: dict, new_units: list[dict]):
            """Looks for the deepest node, i.e. dict in the tree that does not
            have a `units` key.

            After the target is found, it creates a `units` key, pairing
            this key with the value `new_units`
            """
            while True:
                if target := node.get("units", None):
                    node = target[0]
                else:
                    node["units"] = new_units
                    break

        #  first row is root, not necessary for building the branch
        self.nodes.pop(0)
        subtree: dict[str, list[dict] | str] = {}
        while True:
            try:  # for each node that contains units, process
                add_child_units(subtree, [self.nodes.pop(0)])
            except IndexError:
                break
        return subtree.pop("units")  # b/c dict containing a single key


@dataclass
class AuthorDetail:
    email: str
    display: str
    alias: str
    first: str
    last: str
    suffix: str
    org: str
    org_url: str
