import json
import os
import re
from copy import deepcopy
from pathlib import Path
from typing import Any, Iterator, NamedTuple
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

import yaml
from jinja2 import TemplateNotFound

from .etc import (
    JINJA_ENVIRONMENT,
    ResultPagination,
    ResultSetup,
    SearchFilterResults,
    SearchFilterSetup,
    SortFilterSetup,
    get_rows,
    set_result_item,
    set_sql_path,
)

filters_path = (
    Path(os.path.dirname(os.path.realpath(__file__))) / "filters.yaml"
)
"""The filters config file `filters.yaml` must be a sibling file of `search.py`"""

SEARCH_FILTERS = yaml.load(filters_path.read_text(), yaml.SafeLoader)
"""The configuration file for search filters loaded as nested dictionary."""


class SelectedFilter(NamedTuple):
    kind: str
    slug: str


class QueryBase(NamedTuple):
    """Query strings are the basis for search filters, the creation of which is
    based on the fields declared here."""

    q: str
    entity: str
    source: str
    page_num: int | None = 1
    sort_order: str | None = "desc"

    @property
    def partial_format(self):
        """A partial list of query strings without the bits for `source` and
        `entity` (these are supplied later by another function)."""
        return self.make_querystring(
            (
                ("q", self.q),
                ("page_num", self.page_num),
                ("sort_order", self.sort_order),
            )
        )

    @property
    def complete_format(self):
        """When all fields have been populated, the completed data can be
        repurposed to create the query string for the prospective search
        url."""
        return self.make_querystring(
            (
                ("q", self.q),
                ("entity", self.entity),
                ("source", self.source),
                ("page_num", self.page_num),
                ("sort_order", self.sort_order),
            )
        )

    def make_querystring(
        self, pairings: tuple[tuple[str, str | int | None]]
    ) -> str:
        """Adds a `&` to each paired value in encoded in url format, e.g.
        `&q=civil%2Bcode` means q has been assigned the value 'civil code'.

        After all pairings are processed, the initial `&` is removed.
        """
        raw = {k: v for k, v in pairings if v}
        return urlencode(raw).removeprefix("&")

    def get_results(
        self, start_row: int, end_row: int, text: str, url: str
    ) -> ResultSetup:
        """The results are picked based on a specified start row and end row
        (sourced from the `page_num` part of the querystring) that are utilized
        by the `ids.sql`.

        The row numbers are combined with the instantiated `Querybase`
        as a dict to serve as the bind parameters that will be paired
        with the chosen `sqlpath`.
        """
        row_range = {"start_row": start_row, "end_row": end_row}
        bind_params = self._asdict() | row_range
        bind_params["q"] = "'{}'".format(escape_fts(bind_params["q"]))
        sql_stmt = set_sql_path(self.entity, self.source)
        rows = get_rows(sql_stmt, bind_params)
        max_count, results = self.extract(rows)
        return ResultSetup(
            max_count=max_count,
            results=results,
            btn_text=text,
            btn_url=revise_url(url, "page_num", 1),
        )

    def extract(self, rows: Iterator[dict], max_row: int | None = 0) -> tuple:
        search_results = []
        for row in rows:
            # unpopulated parameter in signature;  each row will carry the max_row count, extract this info once
            if "max_row" in row and not max_row:
                max_row = row["max_row"]
            result: dict = {}
            item = set_result_item(row=row, kind=self.entity)
            if row.get("codes"):
                result["codes"] = json.loads(row["codes"])
            search_results.append(result | item)
        return max_row, search_results


class ContextualizeSearchFilters:
    def __init__(self, selection: list[SelectedFilter], url: str) -> None:
        tree = deepcopy(SEARCH_FILTERS)
        self.url = url
        self.selection = selection
        self.add_context(tree)
        self.filters = tree

    def add_context(self, n: dict, url: None = None):
        """Uses the user's selection to add `is_selected` and `url` to the
        filter tree."""
        if n.get("kind") and n.get("slug"):
            partial_querystring = urlencode({n["kind"]: n["slug"]})
            n["url"] = f"{url}&{partial_querystring}"
            for s in self.selection:
                if s == SelectedFilter(n["kind"], n["slug"]):
                    n["is_selected"] = True

        # walk the tree, see FILTERS
        if subnodes := n.get("units", None):
            for sub in subnodes:
                self.add_context(sub, n.get("url", self.url))


class Search:
    def __init__(
        self,
        q: str,
        input_entity: str,
        input_source: str,
        page_num: int | None = None,
        sort_order: str | None = "desc",
        url_base: str | None = "/search?",
        page_size: int = 10,
    ) -> None:
        # determine range of rows based on the page number;
        start = (page_num - 1) * page_size if page_num else 0
        end = start + page_size

        # prepare parameters for searching
        entity, source = self.validate_template(input_entity, input_source)

        # prepare the query string for creating the url and for supplying
        self.querystring = QueryBase(
            q=q,
            entity=entity,
            source=source,
            page_num=page_num,
            sort_order=sort_order,
        )

        full_url = url_base + self.querystring.complete_format
        partial_url = url_base + self.querystring.partial_format

        # query the database, if no page number, assume federated search
        res = self.querystring.get_results(
            start_row=start + 1 if page_num else 1,
            end_row=end if page_num else 3,
            text=f"Search {entity.title()}",
            url=full_url,
        )
        self.max_count = res["max_count"]

        # set pagination urls for related next queries
        if page_num:
            nxt, prv = self.paginate(full_url, end, page_num)
            self.pagination = ResultPagination(
                start_row=start + 1,
                end_row=min([end, self.max_count]),
                next_url=nxt,
                prev_url=prv,
            )

        # formalize selection
        self.selection = [
            SelectedFilter("entity", entity),
            SelectedFilter("source", source),
        ]

        # provide filter urls for other queries
        self.params = ContextualizeSearchFilters(self.selection, partial_url)
        self.keys = ("item", "kind", "slug", "url", "is_selected")
        self.options = self.get_filter_options(self.params.filters)
        self.sorter = self.get_sort_options(full_url, sort_order)

        # finalize context for template
        self.context = self.querystring._asdict() | res
        if page_num:
            self.context = (
                self.context | self.pagination | self.options | self.sorter
            )

    def validate_template(self, cat: str, src: str) -> tuple[str, str]:
        """Is there a valid template, based on the supplied parameters; if
        none, ensure a valid template path string."""
        try:
            JINJA_ENVIRONMENT.get_template(set_sql_path(cat, src))
            return (cat, src)
        except TemplateNotFound:
            if cat == "decisions":
                return (cat, "body_text")
            elif cat == "statutes":
                return (cat, "provisions")
            elif cat == "documents":
                return (cat, "contents")
            else:
                raise Exception(f"Improper {cat=}")

    def paginate(
        self, url: str, end_row: int, page_num: int = 1
    ) -> tuple[str | None, str | None]:
        """Get search results for specific member search having default
        `page_num` of 1, with a `row_count_in_page`, i.e. number of resulting
        items to be displayed within the page, of 10.

        Note nuance on start row for the `data`, the `start` is `end` of previous page, so increment
        - page 1 = start 1, end 10 (next_start = start + 1 so 11)
        - page 3 = start 20, end 30  (next_start = start + 1 so 21)
        """
        self.max_count
        if not self.max_count:  # no results found
            return None, None

        if end_row >= self.max_count:  # num > max, remove nxt, test prv
            nxt, prv = None, None
            if page_num > 1:  # more than the first page, implies prev
                prv = revise_url(url, "page_num", page_num - 1)
        else:  #  row num is within max, show nxt, test prv
            nxt, prv = revise_url(url, "page_num", page_num + 1), None
            if page_num > 1:  # more than the first page, implies prev
                prv = revise_url(url, "page_num", page_num - 1)
        return nxt, prv

    def get_filter_options(self, tree: dict) -> SearchFilterSetup:
        entity_filter = SearchFilterResults(
            item="Entity",
            slug="entity",
            units=(entities := list(self._get_options(tree, "base"))),
        )

        selected_cat = next(c for c in entities if c.get("is_selected"))
        source_filter = SearchFilterResults(
            item="Criteria",
            slug="source",
            units=list(self._get_options(tree, selected_cat["slug"])),
        )

        raw = list(self._get_actives(tree))
        raw.pop(0)  # remove first
        return SearchFilterSetup(
            entities=entity_filter,
            sources=source_filter,
            actives=[f"{a['kind'].capitalize()}: {a['item']}" for a in raw],
        )

    def get_sort_options(self, url: str, sort_order: None | str = "desc"):
        sort_asc = revise_url(url, "sort_order", "asc")
        if sort_order == "asc":
            sort_asc += "&is_selected=True"
        sort_desc = revise_url(url, "sort_order", "desc")
        if sort_order == "desc":
            sort_desc += "&is_selected=True"
        return SortFilterSetup(sort_url_asc=sort_asc, sort_url_desc=sort_desc)

    def _get_options(self, d: dict, selected_slug: str):
        """Each selected slug will contain children nodes."""
        if d.get("slug") and d["slug"] == selected_slug and d.get("units"):
            for node in d["units"]:
                yield {k: v for k, v in node.items() if k in self.keys}
        if d.get("units"):
            for node in d["units"]:
                yield from self._get_options(node, selected_slug)

    def _get_actives(self, d: dict):
        if d.get("is_selected"):
            yield {k: v for k, v in d.items() if k in self.keys}
        if d.get("units"):
            for node in d["units"]:
                yield from self._get_actives(node)


def revise_url(url: str, name: str, new_scalar_value: Any):
    """Replace the url's query string value represented by `name` with the
    `new_scalar_value`.

    See inspiration from https://stackoverflow.com/questions/67192293/delete-only-a-specific-
    query-from-an-url/67192897#67192897.
    """
    url_bits = list(urlparse(url))
    query_string = parse_qs(url_bits[4], keep_blank_values=True)
    if query_string.get(name):
        del query_string[name]
    query_string[name] = new_scalar_value
    url_bits[4] = urlencode(query_string, doseq=True)  # 4th el of bits parsed
    new_url = urlunparse(url_bits)
    return new_url


def escape_fts(query):
    """See https://github.com/simonw/datasette/blob/8e18c7943181f228ce5ebcea48d
    eb59ce50bee1f/datasette/utils/__init__.py#L818-L829."""
    split_pattern = re.compile(r'\s+|(".*?")')
    # e.g. "Republic Act No. 386" becomes a separate match once re.split is employed.

    # If query has unbalanced ", add one at end
    if query.count('"') % 2:
        query += '"'
    bits = split_pattern.split(query)
    bits = [b for b in bits if b and b != '""']
    result = " ".join(
        '"{}"'.format(bit) if not bit.startswith('"') else bit for bit in bits
    )
    return f"{result}"


def double_quote_fts(phrase):
    """When a phrase needs double quotes, it needs to be converted into a
    string literal and then read as double quotes. This function converts
    string into the intended FTS format compliant with
    https://sqlite.org/fts5.html#fts5_strings.

    The first '' pair converts this to a string literal. The the second
    \" \" pairing adds the double quotes required.
    """
    quoted_phrase = f"'\"{phrase}\"'"
    #    f'"{phrase}"'  #
    return quoted_phrase
